from dataclasses import replace
import datetime as dt


import pandas as pd
import pytest
import pytz


from solarforecastarbiter.reference_forecasts import utils
from solarforecastarbiter.conftest import default_forecast, default_observation


@pytest.fixture
def forecast_hr_begin(site_metadata):
    return default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=5),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('1h'),
        run_length=pd.Timedelta('1h'),
        interval_label='beginning')


@pytest.mark.parametrize('issuetime,rl,lt,start,expected', [
    ('06:00', '1h', '1h', pd.Timestamp('20190101T0000Z'),
     [pd.Timestamp('20190101T0000Z')] + list(
         pd.date_range(start='20190101T0600Z', end='20190102T0000Z',
                       freq='1h'))),
    ('00:00', '4h', '1h', pd.Timestamp('20190101T0000Z'),
     [pd.Timestamp('20190101T0000Z'), pd.Timestamp('20190101T0400Z'),
      pd.Timestamp('20190101T0800Z'), pd.Timestamp('20190101T1200Z'),
      pd.Timestamp('20190101T1600Z'), pd.Timestamp('20190101T2000Z'),
      pd.Timestamp('20190102T0000Z')]),
    ('16:00', '8h', '3h', pd.Timestamp('20190101T0000Z'),
     [pd.Timestamp('20190101T0000Z'), pd.Timestamp('20190101T1600Z'),
      pd.Timestamp('20190102T0000Z')]),
    ('00:30', '4h', '120h', pd.Timestamp('20190101T0000Z'),
     [pd.Timestamp('20190101T0030Z'), pd.Timestamp('20190101T0430Z'),
      pd.Timestamp('20190101T0830Z'), pd.Timestamp('20190101T1230Z'),
      pd.Timestamp('20190101T1630Z'), pd.Timestamp('20190101T2030Z'),
      pd.Timestamp('20190102T0030Z')]),
    ('05:00', '6h', '1h', pd.Timestamp('20190101T0000Z'),
     [pd.Timestamp('20190101T0500Z'), pd.Timestamp('20190101T1100Z'),
      pd.Timestamp('20190101T1700Z'), pd.Timestamp('20190101T2300Z'),
      pd.Timestamp('20190102T0500Z')]),
    ('11:00', '12h', '3h', pd.Timestamp('20190101T2300Z'),
     [pd.Timestamp('20190101T1100Z'), pd.Timestamp('20190101T2300Z'),
      pd.Timestamp('20190102T1100Z')]),
    ('05:00', '12h', '1h', pd.Timestamp('20190101T0000Z'),
     [pd.Timestamp('20190101T0500Z'), pd.Timestamp('20190101T1700Z'),
      pd.Timestamp('20190102T0500Z')]),
    ('05:00', '12h', '1h', pd.Timestamp('20190101T0900-06:00'),
     [pd.Timestamp('20190101T1100-06:00'),
      pd.Timestamp('20190101T2300-06:00'), pd.Timestamp('20190102T1100-06:00')
      ]),
    ('05:00', '12h', '1h', pd.Timestamp('20190101T0900'),
     [pd.Timestamp('20190101T0500'), pd.Timestamp('20190101T1700'),
      pd.Timestamp('20190102T0500')]),
    ('10:00', '12h', '1h', pd.Timestamp('20190101T0900-06:00'),
     [pd.Timestamp('20190101T0400-06:00'), pd.Timestamp('20190101T1600-06:00'),
      pd.Timestamp('20190102T0400-06:00')]),
    ('00:00', '6h', '1h', pd.Timestamp('20190101T1801-06:00'),
     [pd.Timestamp('20190101T0000-06:00'), pd.Timestamp('20190101T0600-06:00'),
      pd.Timestamp('20190101T1200-06:00'), pd.Timestamp('20190101T1800-06:00'),
      pd.Timestamp('20190102T0000-06:00')]),
    ('05:00', '24h', '1h', pd.Timestamp('20190101T0900'),
     [pd.Timestamp('20190101T0500'), pd.Timestamp('20190102T0500')]),
    ('05:00', '2d', '1h', pd.Timestamp('20190101T0900'),
     [pd.Timestamp('20190101T0500'), pd.Timestamp('20190103T0500')]),
    ('05:00', '36h', '1h', pd.Timestamp('20190101T0900'),
     [pd.Timestamp('20190101T0500'), pd.Timestamp('20190102T1700')]),
    ('12:00', '7d', '1h', pd.Timestamp('20190101T0900-07:00'),
     [pd.Timestamp('20190101T0500-07:00'),
      pd.Timestamp('20190108T0500-07:00')]),
    ('07:00', '12h', '1h',
     pd.Timestamp('20200307T0100', tz='America/New_York'),
     [pd.Timestamp('20200307T0200', tz='America/New_York'),
      pd.Timestamp('20200307T1400', tz='America/New_York'),
      # daylight savings
      pd.Timestamp('20200308T0300', tz='America/New_York')]
     )
])
def test_issue_times(single_forecast, issuetime, rl, lt, start,
                     expected):
    fx = replace(
        single_forecast,
        issue_time_of_day=dt.datetime.strptime(issuetime, '%H:%M').time(),
        run_length=pd.Timedelta(rl),
        lead_time_to_start=pd.Timedelta(lt))
    out = utils.get_issue_times(fx, start)
    assert out == expected


def test_issue_times_high_freq(single_forecast):
    fx = replace(
        single_forecast,
        issue_time_of_day=dt.time(0),
        run_length=pd.Timedelta('15min'),
        interval_length=pd.Timedelta('5min'),
        lead_time_to_start=pd.Timedelta('5min'))
    out = utils.get_issue_times(fx, pd.Timestamp('20200501T0000-07:00'))
    assert out == list(pd.date_range(
        start='20200501T0000', end='20200502T0000', tz='Etc/GMT+7',
        freq='15min'))


def test_issue_times_high_freq_offset(single_forecast):
    fx = replace(
        single_forecast,
        issue_time_of_day=pytz.timezone('Etc/GMT+7').localize(dt.time(1)),
        run_length=pd.Timedelta('15min'),
        interval_length=pd.Timedelta('5min'),
        lead_time_to_start=pd.Timedelta('5min'))
    out = utils.get_issue_times(fx, pd.Timestamp('20200501T0000-07:00'))
    assert out == [pd.Timestamp('20200501T0000-07:00')] + list(
        pd.date_range(
            start='20200501T0100', end='20200502T0000', tz='Etc/GMT+7',
            freq='15min'))


def test_issue_times_localized(single_forecast):
    fx = replace(
        single_forecast,
        issue_time_of_day=pytz.timezone('Etc/GMT+7').localize(dt.time(12)),
        run_length=pd.Timedelta('12h'),
    )
    out = utils.get_issue_times(fx, pd.Timestamp('20200601T1700-0500'))
    assert out == [pd.Timestamp('20200601T0200-05:00'),
                   pd.Timestamp('20200601T1400-05:00'),
                   pd.Timestamp('20200602T0200-05:00')]


def test_issue_times_localized_dst(single_forecast):
    tzinfo = pd.Timestamp('20200308T0000', tz='America/New_York').tzinfo
    fx = replace(
        single_forecast,
        issue_time_of_day=dt.time(hour=2, tzinfo=tzinfo),
        run_length=pd.Timedelta('12h'),
    )
    out = utils.get_issue_times(fx, pd.Timestamp('20200308T0100-0500'))
    assert out == [pd.Timestamp('20200308T0300-04:00'),
                   pd.Timestamp('20200308T1500-04:00'),
                   pd.Timestamp('20200309T0300-04:00')]


def test_issue_times_fx_gap(forecast_hr_begin):
    # 1-4 utc should not be included in output as 5 is the first
    # issue time for the day.
    out = utils.get_issue_times(forecast_hr_begin,
                                pd.Timestamp('20200505T2000-07:00'))
    assert out == list(pd.date_range(
        start=pd.Timestamp('20200505T0000-07:00'),
        end=pd.Timestamp('20200505T1700-07:00'),
        freq='1h')) + list(pd.date_range(
            start=pd.Timestamp('20200505T2200-07:00'),
            end=pd.Timestamp('20200506T0000-07:00'),
            freq='1h'
        ))


@pytest.mark.parametrize('runtime,expected', [
    (pd.Timestamp('20190501T1100Z'), pd.Timestamp('20190501T1100Z')),
    (pd.Timestamp('20190501T1030Z'), pd.Timestamp('20190501T1100Z')),
    (pd.Timestamp('20190501T0030Z'), pd.Timestamp('20190501T0500Z')),
    (pd.Timestamp('20190501T2359Z'), pd.Timestamp('20190502T0500Z')),
    (pd.Timestamp('20190501T2200Z'), pd.Timestamp('20190501T2300Z')),
    (pd.Timestamp('20190501T0900-07:00'), pd.Timestamp('20190501T1000-07:00'))
])
def test_get_next_issue_time(single_forecast, runtime, expected):
    fx = replace(
        single_forecast,
        issue_time_of_day=dt.time(5, 0),
        run_length=pd.Timedelta('6h'),
        lead_time_to_start=pd.Timedelta('1h'))
    out = utils.get_next_issue_time(fx, runtime)
    assert out == expected


def test_get_init_time():
    run_time = pd.Timestamp('20190501T1200Z')
    fetch_metadata = {'delay_to_first_forecast': '1h',
                      'avg_max_run_length': '5h',
                      'update_freq': '6h'}
    assert utils.get_init_time(run_time, fetch_metadata) == pd.Timestamp(
        '20190501T0600Z')


def test_get_data_start_end_labels_1h_window_limit(site_metadata):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('5min'), interval_label='beginning')
    forecast = default_forecast(
        site_metadata,
        run_length=pd.Timedelta('12h'),  # test 1 hr limit on window
        interval_label='beginning')
    # ensure data no later than run time
    run_time = pd.Timestamp('20190422T1945Z')
    issue_time = pd.Timestamp('20190422T2000Z')
    data_start, data_end = utils.get_data_start_end(observation, forecast,
                                                    run_time, issue_time)
    assert data_start == pd.Timestamp('20190422T1845Z')
    assert data_end == pd.Timestamp('20190422T1945Z')


def test_get_data_start_end_labels_subhourly_window_limit(site_metadata):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('5min'), interval_label='beginning')
    forecast = default_forecast(
        site_metadata,
        run_length=pd.Timedelta('5min'),  # test subhourly limit on window
        interval_label='beginning')
    run_time = pd.Timestamp('20190422T1945Z')
    issue_time = pd.Timestamp('20190422T2000Z')
    data_start, data_end = utils.get_data_start_end(observation, forecast,
                                                    run_time, issue_time)
    assert data_start == pd.Timestamp('20190422T1940Z')
    assert data_end == pd.Timestamp('20190422T1945Z')


def test_get_data_start_end_labels_obs_longer_than_fx_run(site_metadata):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('15min'))
    forecast = default_forecast(
        site_metadata,
        run_length=pd.Timedelta('5min'))
    run_time = pd.Timestamp('20190422T1945Z')
    issue_time = pd.Timestamp('20190422T2000Z')
    # obs interval cannot be longer than forecast interval
    with pytest.raises(ValueError) as excinfo:
        utils.get_data_start_end(observation, forecast, run_time, issue_time)
    assert ("observation.interval_length <= forecast.run_length" in
            str(excinfo.value))


def test_get_data_start_end_labels_obs_longer_than_1h(site_metadata):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('2h'))
    forecast = default_forecast(
        site_metadata,
        run_length=pd.Timedelta('5min'))
    run_time = pd.Timestamp('20190422T1945Z')
    issue_time = pd.Timestamp('20190422T2000Z')
    # obs interval cannot be longer than 1 hr
    with pytest.raises(ValueError) as excinfo:
        utils.get_data_start_end(observation, forecast, run_time, issue_time)
    assert 'observation.interval_length <= 1h' in str(excinfo.value)


def test_get_data_start_end_labels_obs_longer_than_1h_day_ahead(site_metadata):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('2h'), interval_label='beginning')
    forecast = default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=5),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('1h'),
        run_length=pd.Timedelta('1d'),  # day ahead
        interval_label='beginning')
    run_time = pd.Timestamp('20190422T1945Z')
    issue_time = pd.Timestamp('20190423T0500Z')
    # day ahead doesn't care about obs interval length
    data_start, data_end = utils.get_data_start_end(observation, forecast,
                                                    run_time, issue_time)
    assert data_start == pd.Timestamp('20190421T0600Z')
    assert data_end == pd.Timestamp('20190422T0600Z')


def test_get_data_start_end_labels_obs_fx_instant_mismatch(site_metadata):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('5min'), interval_label='instant')
    forecast = default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=5),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('1h'),  # interval_length must be equal
        run_length=pd.Timedelta('1d'),
        interval_label='instant')            # if interval_label also instant
    run_time = pd.Timestamp('20190422T1945Z')
    issue_time = pd.Timestamp('20190422T2000Z')
    with pytest.raises(ValueError) as excinfo:
        utils.get_data_start_end(observation, forecast, run_time, issue_time)
    assert 'with identical interval length' in str(excinfo.value)


@pytest.mark.parametrize('it,lead,issue', [
    (23, '1h', '20190422T2300Z'),
    (5, '19h', '20190422T0500Z'),
    (5, '19h', '20190422T0000-05:00')
])
def test_get_data_start_end_labels_obs_fx_instant(site_metadata, lead, issue,
                                                  it):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('5min'), interval_label='instant')
    # interval length of forecast and obs must be equal if interval label is
    # instant
    forecast = default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=it),
        lead_time_to_start=pd.Timedelta(lead),
        interval_length=pd.Timedelta('5min'),
        run_length=pd.Timedelta('1d'),
        interval_label='instant')
    issue_time = pd.Timestamp(issue)
    run_time = issue_time - pd.Timedelta('75min')
    data_start, data_end = utils.get_data_start_end(observation, forecast,
                                                    run_time, issue_time)
    assert data_start == pd.Timestamp('20190421T0000Z')
    assert data_end == pd.Timestamp('20190421T235959Z')


def test_get_data_start_end_labels_obs_instant_fx_avg(site_metadata):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('5min'), interval_label='instant')
    forecast = default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=23),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('5min'),
        run_length=pd.Timedelta('1d'),
        interval_label='beginning')
    run_time = pd.Timestamp('20190422T1945Z')
    issue_time = pd.Timestamp('20190422T2300Z')
    data_start, data_end = utils.get_data_start_end(observation, forecast,
                                                    run_time, issue_time)
    assert data_start == pd.Timestamp('20190421T0000Z')
    assert data_end == pd.Timestamp('20190421T235959Z')


def test_get_data_start_end_labels_obs_instant_fx_avg_ending(site_metadata):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('5min'), interval_label='instant')
    forecast = default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=0),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('5min'),
        run_length=pd.Timedelta('1d'),
        interval_label='ending')
    run_time = pd.Timestamp('20190422T1945Z')
    issue_time = pd.Timestamp('20190422T2300Z')
    data_start, data_end = utils.get_data_start_end(observation, forecast,
                                                    run_time, issue_time)
    assert data_start == pd.Timestamp('20190421T000001Z')
    assert data_end == pd.Timestamp('20190422T0000Z')


def test_get_data_start_end_labels_obs_instant_fx_avg_intraday(site_metadata):
    run_time = pd.Timestamp('20190422T1945Z')
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('5min'), interval_label='instant')
    forecast = default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=5),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('5min'),
        run_length=pd.Timedelta('15min'),
        interval_label='ending')
    issue_time = pd.Timestamp('20190422T2000Z')
    data_start, data_end = utils.get_data_start_end(observation, forecast,
                                                    run_time, issue_time)
    assert data_start == pd.Timestamp('20190422T193001Z')
    assert data_end == pd.Timestamp('20190422T1945Z')


def test_get_data_start_end_labels_obs_avg_fx_instant(site_metadata):
    run_time = pd.Timestamp('20190422T1945Z')
    issue_time = pd.Timestamp('20190422T2000Z')
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('5min'), interval_label='ending')
    forecast = default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=5),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('5min'),
        run_length=pd.Timedelta('1d'),
        interval_label='instant')
    with pytest.raises(ValueError) as excinfo:
        utils.get_data_start_end(observation, forecast, run_time, issue_time)
    assert 'made from interval average obs' in str(excinfo.value)


@pytest.mark.parametrize('rl,expected_start,expected_end,rt,lt', [
    ('1d', '20190409T0000Z', '20190410T0000Z', '20190410T2230Z', '1h'),
    ('1d', '20190409T0000Z', '20190410T0000Z', '20190410T0001Z', '1h'),
    ('1d', '20190409T0000Z', '20190410T0000Z', '20190410T0001Z', '25h'),
    ('1d', '20190409T0000Z', '20190410T0000Z', '20190410T0001Z', '49h'),
    ('1d', '20190408T2300Z', '20190409T2300Z', '20190410T2230Z', '24h'),
    # quirky that 23h ahead uses more recent data than 1h ahead, but
    # timestamps are maintained and end is before issue time
    ('1d', '20190409T2200Z', '20190410T2200Z', '20190410T2230Z', '23h'),
    ('2d', '20190408T0000Z', '20190410T0000Z', '20190410T2230Z', '1h'),
    ('36h', '20190409T0000Z', '20190410T1200Z', '20190410T2300Z', '1h'),
    # run time is too early compared to issue time so data_end > run_time
    # and that's ok.
    ('36h', '20190409T0000Z', '20190410T1200Z', '20190410T1100Z', '1h'),
    ('36h', '20190409T0000Z', '20190410T1200Z', '20190410T1100Z', '25h')
])
def test_get_data_start_end_time_dayahead(site_metadata, rl, rt, lt,
                                          expected_start, expected_end):
    observation = default_observation(
        site_metadata,
        interval_length=pd.Timedelta('5min'), interval_label='beginning'
    )

    run_time = pd.Timestamp(rt)
    issue_time = pd.Timestamp('20190410T2300Z')
    forecast = default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=23),
        lead_time_to_start=pd.Timedelta(lt),
        interval_length=pd.Timedelta('1h'),
        run_length=pd.Timedelta(rl),
        interval_label='beginning')
    data_start, data_end = utils.get_data_start_end(
        observation, forecast, run_time, issue_time)
    assert data_start == pd.Timestamp(expected_start)
    assert data_end == pd.Timestamp(expected_end)


@pytest.mark.parametrize("variable,expected_start,expected_end,rl", [
    # generate forecast on Wednesday 4/10 that applies to Thursday 4/11 using
    # data from the previous Thursday (4/4)
    ("net_load", "20190404T0000Z", "20190405T0000Z", "1d"),
    ("net_load", "20190404T0000Z", "20190407T0000Z", "3d"),
    # generate forecast on Wednesday 4/10 that applies to Thursday 4/11 using
    # data from the previous day (Tuesday 4/9)
    ("ghi", "20190409T0000Z", "20190410T0000Z", "1d"),
])
def test_get_data_start_end_time_weekahead(site_metadata, variable, rl,
                                           expected_start, expected_end):
    observation = default_observation(
        site_metadata, variable=variable,
        interval_length=pd.Timedelta('5min'), interval_label='beginning'
    )

    run_time = pd.Timestamp('20190410T2230Z')
    issue_time = pd.Timestamp('20190410T2300Z')
    # fx from 2019-04-11 00:00
    forecast = default_forecast(
        site_metadata, variable=variable,
        issue_time_of_day=dt.time(hour=23),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('1h'),
        run_length=pd.Timedelta(rl),
        interval_label='beginning')
    data_start, data_end = utils.get_data_start_end(
        observation, forecast, run_time, issue_time)
    assert data_start == pd.Timestamp(expected_start)
    assert data_end == pd.Timestamp(expected_end)


@pytest.mark.parametrize("variable,expected_start,expected_end,rl,issue,run", [
    ("net_load", "20190404T0000Z", "20190405T0000Z", "1d",
     '20190410T1600-07:00', '20190410T1550-07:00'),
    ("net_load", "20190410T2150Z", "20190410T2250Z", "12h",
     '20190410T1600-07:00', '20190410T1550-07:00'),
    ("ghi", "20190404T0000Z", "20190405T0000Z", "1d",
     '20190405T1600-07:00', '20190405T1550-07:00'),
    ("ghi", "20190404T0000Z", "20190405T0000Z", "1d",
     '20190405T1600-07:00', '20190405T1750-05:00'),
    ("ghi", "20190405T1450-07:00", "20190405T1550-07:00", "8h",
     '20190405T1600-07:00', '20190405T1550-07:00'),
])
def test_get_data_start_end_time_tz(site_metadata, variable, rl, issue, run,
                                    expected_start, expected_end):
    observation = default_observation(
        site_metadata, variable=variable,
        interval_length=pd.Timedelta('5min'), interval_label='ending'
    )
    forecast = default_forecast(
        site_metadata, variable=variable,
        issue_time_of_day=dt.time(hour=23),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('1h'),
        run_length=pd.Timedelta(rl),
        interval_label='beginning')
    data_start, data_end = utils.get_data_start_end(
        observation, forecast, pd.Timestamp(run), pd.Timestamp(issue))
    assert data_start == pd.Timestamp(expected_start)
    assert data_end == pd.Timestamp(expected_end)


def test_get_data_start_end_time_weekahead_not_midnight(site_metadata):
    variable = 'net_load'
    observation = default_observation(
        site_metadata, variable=variable,
        interval_length=pd.Timedelta('5min'), interval_label='beginning'
    )

    run_time = pd.Timestamp('20190410T1030Z')
    issue_time = pd.Timestamp('20190410T1200Z')
    # fx from 2019-04-11 12:00
    forecast = default_forecast(
        site_metadata, variable=variable,
        issue_time_of_day=dt.time(hour=12),
        lead_time_to_start=pd.Timedelta('1d'),
        interval_length=pd.Timedelta('1h'),
        run_length=pd.Timedelta('1d'),
        interval_label='beginning')
    data_start, data_end = utils.get_data_start_end(
        observation, forecast, run_time, issue_time)
    assert data_start == pd.Timestamp('20190404T1200Z')
    assert data_end == pd.Timestamp('20190405T1200Z')


@pytest.mark.parametrize('adjust_for_interval_label', [True, False])
def test_get_forecast_start_end_time_same_as_issue_time(
        forecast_hr_begin, adjust_for_interval_label):
    # time is same as issue time of day
    issue_time = pd.Timestamp('20190422T0500')
    fx_start, fx_end = utils.get_forecast_start_end(
        forecast_hr_begin, issue_time, adjust_for_interval_label)
    assert fx_start == pd.Timestamp('20190422T0600')
    fx_end_exp = pd.Timestamp('20190422T0700')
    if adjust_for_interval_label:
        fx_end_exp -= pd.Timedelta('1n')
    assert fx_end == fx_end_exp


@pytest.mark.parametrize('adjust_for_interval_label', [True, False])
@pytest.mark.parametrize('issue,exp_start,exp_end', [
    # issue time at 5, 1hr lead/interval
    (pd.Timestamp('20190422T0500'), pd.Timestamp('20190422T0600'),
     pd.Timestamp('20190422T0700')),
    (pd.Timestamp('20190422T0500Z'), pd.Timestamp('20190422T0600Z'),
     pd.Timestamp('20190422T0700Z')),
    (pd.Timestamp('20190422T0000-05:00'), pd.Timestamp('20190422T0100-05:00'),
     pd.Timestamp('20190422T0200-05:00')),
    (pd.Timestamp('20190421T2300-06:00'), pd.Timestamp('20190422T0000-06:00'),
     pd.Timestamp('20190422T0100-06:00')),
    (pd.Timestamp('20190422T0700-07:00'), pd.Timestamp('20190422T0800-07:00'),
     pd.Timestamp('20190422T0900-07:00')),
    # there is a gap in forecasts since issue time is 5 and only hourly run
    pytest.param(pd.Timestamp('20190422T2000-07:00'),  # 3 utc
                 pd.Timestamp('20190422T2100-07:00'),
                 pd.Timestamp('20190422T2200-07:00'),
                 marks=pytest.mark.xfail(strict=True)),
])
def test_get_forecast_start_end_time_tz(
        forecast_hr_begin, adjust_for_interval_label, issue,
        exp_start, exp_end):
    fx_start, fx_end = utils.get_forecast_start_end(
        forecast_hr_begin, issue, adjust_for_interval_label)
    assert fx_start == exp_start
    fx_end_exp = exp_end
    if adjust_for_interval_label:
        fx_end_exp -= pd.Timedelta('1n')
    assert fx_end == fx_end_exp


@pytest.mark.parametrize('adjust_for_interval_label', [True, False])
def test_get_forecast_start_end_time_same_as_issue_time_n_x_run(
        forecast_hr_begin, adjust_for_interval_label):
    # time is same as issue time of day + n * run_length
    issue_time = pd.Timestamp('20190422T1200')
    fx_start, fx_end = utils.get_forecast_start_end(
        forecast_hr_begin, issue_time, adjust_for_interval_label)
    assert fx_start == pd.Timestamp('20190422T1300')
    fx_end_exp = pd.Timestamp('20190422T1400')
    if adjust_for_interval_label:
        fx_end_exp -= pd.Timedelta('1n')
    assert fx_end == fx_end_exp


def test_get_forecast_start_end_time_before_issue_time(forecast_hr_begin):
    # time is before issue time of day but otherwise valid
    issue_time = pd.Timestamp('20190422T0400')
    with pytest.raises(ValueError):
        fx_start, fx_end = utils.get_forecast_start_end(
            forecast_hr_begin, issue_time)


def test_get_forecast_start_end_time_invalid(forecast_hr_begin):
    # time is invalid
    issue_time = pd.Timestamp('20190423T0505')
    with pytest.raises(ValueError):
        fx_start, fx_end = utils.get_forecast_start_end(
            forecast_hr_begin, issue_time)


@pytest.mark.parametrize('adjust_for_interval_label', [True, False])
def test_get_forecast_start_end_time_instant(
        site_metadata, adjust_for_interval_label):
    # instant
    forecast = default_forecast(
        site_metadata,
        issue_time_of_day=dt.time(hour=5),
        lead_time_to_start=pd.Timedelta('1h'),
        interval_length=pd.Timedelta('5min'),
        run_length=pd.Timedelta('1h'),
        interval_label='instant')
    issue_time = pd.Timestamp('20190422T0500')
    fx_start, fx_end = utils.get_forecast_start_end(
        forecast, issue_time, adjust_for_interval_label)
    assert fx_start == pd.Timestamp('20190422T0600')
    fx_end_exp = pd.Timestamp('20190422T0700')
    if adjust_for_interval_label:
        fx_end_exp -= pd.Timedelta('1n')
    assert fx_end == fx_end_exp


@pytest.mark.parametrize('fxargs,last_time,expected', [
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('0h'),
          run_length='1h',
          interval_label='beginning'),
     pd.Timestamp('2020-05-20T10:00Z'),
     pd.Timestamp('2020-05-20T11:00Z')),
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('0h'),
          run_length='1h',
          interval_label='beginning'),
     pd.Timestamp('2020-05-20T10:00-07:00'),
     pd.Timestamp('2020-05-20T11:00-07:00')),
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('0h'),
          run_length='1h',
          interval_label='beginning'),
     pd.Timestamp('2020-05-20T10:30Z'),  # inconsistent w/ fx
     pd.Timestamp('2020-05-20T11:00Z')),
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('0h'),
          run_length='1h',
          interval_label='ending'),
     pd.Timestamp('2020-05-20T10:00Z'),
     pd.Timestamp('2020-05-20T10:00Z')),
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('1h'),
          run_length='1h',
          interval_label='ending'),
     pd.Timestamp('2020-05-20T10:00Z'),
     pd.Timestamp('2020-05-20T09:00Z')),
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('1h'),
          run_length='1h',
          interval_label='beginning'),
     pd.Timestamp('2020-05-20T10:00Z'),
     pd.Timestamp('2020-05-20T10:00Z')),
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('2h'),
          run_length='4h',
          interval_label='beginning'),
     pd.Timestamp('2020-05-20T05:00Z'),
     pd.Timestamp('2020-05-20T04:00Z')),
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('2h'),
          run_length='4h',
          interval_label='ending'),
     pd.Timestamp('2020-05-20T06:00Z'),
     pd.Timestamp('2020-05-20T04:00Z')),
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('2h'),
          run_length='4h',
          interval_label='ending'),
     pd.Timestamp('2020-05-20T05:00Z'),  # run not finished?
     pd.Timestamp('2020-05-20T00:00Z')),
    (dict(issue_time_of_day=dt.time(hour=0),
          lead_time_to_start=pd.Timedelta('2h'),
          run_length='4h',
          interval_label='ending'),
     pd.Timestamp('2020-05-20T06:00:01Z'),  # run not finished?
     pd.Timestamp('2020-05-20T04:00Z')),
    (dict(issue_time_of_day=dt.time(hour=6),
          lead_time_to_start=pd.Timedelta('24h'),
          run_length='8h',
          interval_label='instantaneous'),
     pd.Timestamp('2020-05-20T04:00Z'),
     pd.Timestamp('2020-05-18T22:00Z')),
])
def test_find_next_issue_time_from_last_forecast(fxargs, last_time, expected,
                                                 site_metadata):
    fx = default_forecast(site_metadata, **fxargs)
    out = utils.find_next_issue_time_from_last_forecast(fx, last_time)
    assert out == expected


@pytest.mark.parametrize('obs_kw,fx_kw,index', [
    pytest.param(dict(), dict(), False,
                 marks=pytest.mark.xfail(strict=True)),
    pytest.param(dict(), dict(), True,
                 marks=pytest.mark.xfail(strict=True)),
    # obs interval length > 1h too long
    (dict(interval_length=pd.Timedelta('2h')),
     dict(interval_length=pd.Timedelta('4h')), True),
    (dict(interval_length=pd.Timedelta('2h')),
     dict(interval_length=pd.Timedelta('4h')), False),
    # fx run_length < obs interval length
    (dict(), dict(run_length=pd.Timedelta('15min')), True),
    (dict(), dict(run_length=pd.Timedelta('15min')), False),
    # day forecast index
    (dict(), dict(run_length=pd.Timedelta('24h')), True),
    pytest.param(dict(), dict(run_length=pd.Timedelta('24h')), False,
                 marks=pytest.mark.xfail(strict=True)),
    # non-instant obs
    (dict(), dict(interval_label='instantaneous'), True),
    (dict(), dict(interval_label='instantaneous'), False),
    # instant, but mismatch length
    (dict(interval_label='instantaneous'),
     dict(interval_label='instantaneous'),
     True),
    (dict(interval_label='instantaneous'),
     dict(interval_label='instantaneous'),
     False),
])
def test_check_persistence_compatibility(obs_kw, fx_kw, index, site_metadata):
    obs_dict = {'interval_label': 'ending',
                'interval_value_type': 'interval_mean',
                'interval_length': pd.Timedelta('30min')}
    fx_dict = {'interval_label': 'ending',
               'interval_value_type': 'interval_mean',
               'interval_length': pd.Timedelta('1h'),
               'run_length': pd.Timedelta('12h')}
    obs_dict.update(obs_kw)
    fx_dict.update(fx_kw)
    obs = default_observation(site_metadata, **obs_dict)
    fx = default_forecast(site_metadata, **fx_dict)
    with pytest.raises(ValueError):
        utils.check_persistence_compatibility(obs, fx, index)



"""
Created on Fri Feb 15 14:08:20 2019

@author: cwhanse
"""
import warnings


import numpy as np
import pandas as pd
from pvlib.tools import cosd
from pvlib.irradiance import clearsky_index
from pvlib.clearsky import detect_clearsky as _detect_clearsky


from solarforecastarbiter.validation.quality_mapping import mask_flags


QCRAD_LIMITS = {'ghi_ub': {'mult': 1.5, 'exp': 1.2, 'min': 100},
                'dhi_ub': {'mult': 0.95, 'exp': 1.2, 'min': 50},
                'dni_ub': {'mult': 1.0, 'exp': 0.0, 'min': 0},
                'ghi_lb': -4, 'dhi_lb': -4, 'dni_lb': -4}

QCRAD_CONSISTENCY = {
    'ghi_ratio':
        {'low_zenith':
            {'zenith_bounds': [0.0, 75], 'ghi_bounds': [50, np.Inf],
             'ratio_bounds': [0.92, 1.08]},
         'high_zenith':
            {'zenith_bounds': [75, 93], 'ghi_bounds': [50, np.Inf],
             'ratio_bounds': [0.85, 1.15]}},
    'dhi_ratio':
        {'low_zenith':
            {'zenith_bounds': [0.0, 75], 'ghi_bounds': [50, np.Inf],
             'ratio_bounds': [0.0, 1.05]},
         'high_zenith':
            {'zenith_bounds': [75, 93], 'ghi_bounds': [50, np.Inf],
             'ratio_bounds': [0.0, 1.10]}}}


def _check_limits(val, lb=None, ub=None, lb_ge=False, ub_le=False):
    """ Returns True where lb < (or <=) val < (or <=) ub
    """
    if lb_ge:
        lb_op = np.greater_equal
    else:
        lb_op = np.greater
    if ub_le:
        ub_op = np.less_equal
    else:
        ub_op = np.less

    if (lb is not None) & (ub is not None):
        return lb_op(val, lb) & ub_op(val, ub)
    elif lb is not None:
        return lb_op(val, lb)
    elif ub is not None:
        return ub_op(val, ub)
    else:
        raise ValueError('must provide either upper or lower bound')


def _QCRad_ub(dni_extra, sza, lim):
    cosd_sza = cosd(sza)
    cosd_sza[cosd_sza < 0] = 0
    return lim['mult'] * dni_extra * cosd_sza**lim['exp'] + lim['min']


@mask_flags('LIMITS EXCEEDED')
def check_ghi_limits_QCRad(ghi, solar_zenith, dni_extra, limits=None):
    """
    Tests for physical limits on GHI using the QCRad criteria.

    Test passes if a value > lower bound and value < upper bound. Lower bounds
    are constant for all tests. Upper bounds are calculated as

    .. math::
        ub = min + mult * dni_extra * cos( solar_zenith)^exp

    Parameters
    ----------
    ghi : Series
        Global horizontal irradiance in W/m^2
    solar_zenith : Series
        Solar zenith angle in degrees
    dni_extra : Series
        Extraterrestrial normal irradiance in W/m^2
    limits : dict, default QCRAD_LIMITS
        for keys 'ghi_ub', 'dhi_ub', 'dni_ub', value is a dict with
        keys {'mult', 'exp', 'min'}. For keys 'ghi_lb', 'dhi_lb', 'dni_lb',
        value is a float.

    Returns
    -------
    ghi_limit_flag : Series
        True if value passes physically-possible test
    """
    if not limits:
        limits = QCRAD_LIMITS
    ghi_ub = _QCRad_ub(dni_extra, solar_zenith, limits['ghi_ub'])

    ghi_limit_flag = _check_limits(ghi, limits['ghi_lb'], ghi_ub)
    ghi_limit_flag.name = 'ghi_limit_flag'

    return ghi_limit_flag


@mask_flags('LIMITS EXCEEDED')
def check_dhi_limits_QCRad(dhi, solar_zenith, dni_extra, limits=None):
    """
    Tests for physical limits on DHI using the QCRad criteria.

    Test passes if a value > lower bound and value < upper bound. Lower bounds
    are constant for all tests. Upper bounds are calculated as

    .. math::
        ub = min + mult * dni_extra * cos( solar_zenith)^exp

    Parameters
    ----------
    dhi : Series
        Diffuse horizontal irradiance in W/m^2
    solar_zenith : Series
        Solar zenith angle in degrees
    dni_extra : Series
        Extraterrestrial normal irradiance in W/m^2
    limits : dict, default QCRAD_LIMITS
        for keys 'ghi_ub', 'dhi_ub', 'dni_ub', value is a dict with
        keys {'mult', 'exp', 'min'}. For keys 'ghi_lb', 'dhi_lb', 'dni_lb',
        value is a float.

    Returns
    -------
    dhi_limit_flag : Series
        True if value passes physically-possible test
    """
    if not limits:
        limits = QCRAD_LIMITS

    dhi_ub = _QCRad_ub(dni_extra, solar_zenith, limits['dhi_ub'])

    dhi_limit_flag = _check_limits(dhi, limits['dhi_lb'], dhi_ub)
    dhi_limit_flag.name = 'dhi_limit_flag'

    return dhi_limit_flag


@mask_flags('LIMITS EXCEEDED')
def check_dni_limits_QCRad(dni, solar_zenith, dni_extra, limits=None):
    """
    Tests for physical limits on DNI using the QCRad criteria.

    Test passes if a value > lower bound and value < upper bound. Lower bounds
    are constant for all tests. Upper bounds are calculated as

    .. math::
        ub = min + mult * dni_extra * cos( solar_zenith)^exp

    Parameters
    ----------
    dni : Series
        Direct normal irradiance in W/m^2
    solar_zenith : Series
        Solar zenith angle in degrees
    dni_extra : Series
        Extraterrestrial normal irradiance in W/m^2
    limits : dict, default QCRAD_LIMITS
        for keys 'ghi_ub', 'dhi_ub', 'dni_ub', value is a dict with
        keys {'mult', 'exp', 'min'}. For keys 'ghi_lb', 'dhi_lb', 'dni_lb',
        value is a float.

    Returns
    -------
    dni_limit_flag : Series
        True if value passes physically-possible test
    """
    if not limits:
        limits = QCRAD_LIMITS

    dni_ub = _QCRad_ub(dni_extra, solar_zenith, limits['dni_ub'])

    dni_limit_flag = _check_limits(dni, limits['dni_lb'], dni_ub)
    dni_limit_flag.name = 'dni_limit_flag'

    return dni_limit_flag


@mask_flags('LIMITS EXCEEDED')
def check_irradiance_limits_QCRad(solar_zenith, dni_extra, ghi=None, dhi=None,
                                  dni=None, limits=None):
    """
    Tests for physical limits on GHI, DHI or DNI using the QCRad criteria.

    Test passes if a value > lower bound and value < upper bound. Lower bounds
    are constant for all tests. Upper bounds are calculated as

    .. math::
        ub = min + mult * dni_extra * cos( solar_zenith)^exp

    Parameters
    ----------
    solar_zenith : Series
        Solar zenith angle in degrees
    dni_extra : Series
        Extraterrestrial normal irradiance in W/m^2
    ghi : Series or None, default None
        Global horizontal irradiance in W/m^2
    dhi : Series or None, default None
        Diffuse horizontal irradiance in W/m^2
    dni : Series or None, default None
        Direct normal irradiance in W/m^2
    limits : dict, default QCRAD_LIMITS
        for keys 'ghi_ub', 'dhi_ub', 'dni_ub', value is a dict with
        keys {'mult', 'exp', 'min'}. For keys 'ghi_lb', 'dhi_lb', 'dni_lb',
        value is a float.

    Returns
    -------
    ghi_limit_flag : Series or None, default None
        True if value passes physically-possible test
    dhi_limit_flag : Series or None, default None
    dhi_limit_flag : Series or None, default None

    References
    ----------
    [1] C. N. Long and Y. Shi, An Automated Quality Assessment and Control
        Algorithm for Surface Radiation Measurements, The Open Atmospheric
        Science Journal 2, pp. 23-37, 2008.
    """
    if not limits:
        limits = QCRAD_LIMITS

    if ghi is not None:
        ghi_limit_flag = check_ghi_limits_QCRad(ghi, solar_zenith, dni_extra,
                                                limits=limits)
    else:
        ghi_limit_flag = None

    if dhi is not None:
        dhi_limit_flag = check_dhi_limits_QCRad(dhi, solar_zenith, dni_extra,
                                                limits=limits)
    else:
        dhi_limit_flag = None

    if dni is not None:
        dni_limit_flag = check_dni_limits_QCRad(dni, solar_zenith, dni_extra,
                                                limits=limits)
    else:
        dni_limit_flag = None

    return ghi_limit_flag, dhi_limit_flag, dni_limit_flag


def _get_bounds(bounds):
    return (bounds['ghi_bounds'][0], bounds['ghi_bounds'][1],
            bounds['zenith_bounds'][0], bounds['zenith_bounds'][1],
            bounds['ratio_bounds'][0], bounds['ratio_bounds'][1])


def _check_irrad_ratio(ratio, ghi, sza, bounds):
    # unpack bounds dict
    ghi_lb, ghi_ub, sza_lb, sza_ub, ratio_lb, ratio_ub = _get_bounds(bounds)
    # for zenith set lb_ge to handle edge cases, e.g., zenith=0
    return ((_check_limits(sza, lb=sza_lb, ub=sza_ub, lb_ge=True)) &
            (_check_limits(ghi, lb=ghi_lb, ub=ghi_ub)) &
            (_check_limits(ratio, lb=ratio_lb, ub=ratio_ub)))


@mask_flags('INCONSISTENT IRRADIANCE COMPONENTS')
def check_irradiance_consistency_QCRad(ghi, solar_zenith, dni_extra, dhi, dni,
                                       param=None):
    """
    Checks consistency of GHI, DHI and DNI. Not valid for night time.

    Parameters
    ----------
    ghi : Series
        Global horizontal irradiance in W/m^2
    solar_zenith : Series
        Solar zenith angle in degrees
    dni_extra : Series
        Extraterrestrial normal irradiance in W/m^2
    dhi : Series
        Diffuse horizontal irradiance in W/m^2
    dni : Series
        Direct normal irradiance in W/m^2
    param : dict
        keys are 'ghi_ratio' and 'dhi_ratio'. For each key, value is a dict
        with keys 'high_zenith' and 'low_zenith'; for each of these keys,
        value is a dict with keys 'zenith_bounds', 'ghi_bounds', and
        'ratio_bounds' and value is an ordered pair [lower, upper]
        of float.

    Returns
    -------
    consistent_components : Series
        True if ghi, dhi and dni components are consistent.
    diffuse_ratio_limit : Series
        True if diffuse to ghi ratio passes limit test.

    References
    ----------
    [1] C. N. Long and Y. Shi, An Automated Quality Assessment and Control
        Algorithm for Surface Radiation Measurements, The Open Atmospheric
        Science Journal 2, pp. 23-37, 2008.
    """

    if not param:
        param = QCRAD_CONSISTENCY

    # sum of components
    component_sum = dni * cosd(solar_zenith) + dhi
    ghi_ratio = ghi / component_sum
    dhi_ratio = dhi / ghi

    bounds = param['ghi_ratio']
    consistent_components = (
        _check_irrad_ratio(ratio=ghi_ratio, ghi=component_sum,
                           sza=solar_zenith, bounds=bounds['high_zenith']) |
        _check_irrad_ratio(ratio=ghi_ratio, ghi=component_sum,
                           sza=solar_zenith, bounds=bounds['low_zenith']))
    consistent_components.name = 'consistent_components'

    bounds = param['dhi_ratio']
    diffuse_ratio_limit = (
        _check_irrad_ratio(ratio=dhi_ratio, ghi=ghi, sza=solar_zenith,
                           bounds=bounds['high_zenith']) |
        _check_irrad_ratio(ratio=dhi_ratio, ghi=ghi, sza=solar_zenith,
                           bounds=bounds['low_zenith']))
    diffuse_ratio_limit.name = 'diffuse_ratio_limit'

    return consistent_components, diffuse_ratio_limit


@mask_flags('LIMITS EXCEEDED')
def check_temperature_limits(temp_air, temp_limits=(-35., 50.)):
    """ Checks for extreme temperatures.

    Parameters
    ----------
    temp_air : Series
        Air temperature in Celsius
    temp_limits : tuple, default (-35, 50)
        (lower bound, upper bound) for temperature.

    Returns
    -------
    extreme_temp_flag : Series
        True if temp_air > lower bound and temp_air < upper bound.
    """
    extreme_temp_flag = _check_limits(temp_air, lb=temp_limits[0],
                                      ub=temp_limits[1])
    extreme_temp_flag.name = 'extreme_temp_flag'
    return extreme_temp_flag


@mask_flags('LIMITS EXCEEDED')
def check_wind_limits(wind_speed, wind_limits=(0., 50.)):
    """ Checks for extreme wind speeds.

    Parameters
    ----------
    wind_speed : Series
        Wind speed in m/s
    wind_limits : tuple, default (0, 50)
        (lower bound, upper bound) for wind speed.

    Returns
    -------
    extreme_wind_flag : Series
        True if wind_speed > lower bound and wind_speed < upper bound.
    """
    extreme_wind_flag = _check_limits(wind_speed, lb=wind_limits[0],
                                      ub=wind_limits[1],
                                      lb_ge=True)
    extreme_wind_flag.name = 'extreme_wind_flag'
    return extreme_wind_flag


@mask_flags('LIMITS EXCEEDED')
def check_rh_limits(rh, rh_limits=(0, 100)):
    """ Checks for extreme relative humidity.

    Parameters
    ----------
    rh : Series
        Relative humidity in %
    rh_limits : tuple, default (0, 100)
        (lower bound, upper bound) for relative humidity

    Returns
    -------
    flags : Series
        True if rh >= lower bound and rh <= upper bound.
    """
    flags = _check_limits(rh, lb=rh_limits[0], ub=rh_limits[1], lb_ge=True,
                          ub_le=True)
    return flags


@mask_flags('LIMITS EXCEEDED')
def check_ac_power_limits(power, solar_zenith, capacity,
                          capacity_limit_low=-0.05,
                          capacity_limit_high_day=1.05,
                          capacity_limit_high_night=0.05):
    """ Checks for extreme AC power.

    Parameters
    ----------
    power : Series
        DC or AC power.
    solar_zenith : Series
        Solar zenith angle in degrees.
    capacity : float
        AC capacity.
    capacity_limit_low : float
        Lower bound in fraction of capacity.
    capacity_limit_high_day : float
        Upper bound in fraction of capacity for day time.
    capacity_limit_high_night : float
        Upper bound in fraction of capacity for night time.

    Returns
    -------
    flags : Series
        True for values that are within the limits:
          * power > capacity * capacity_limit_low, AND
            * power < capacity * capacity_limit_high_day AND solar_zenith < 93, OR
            * power < capacity * capacity_limit_high_night AND solar_zenith > 93

    Notes
    -----
    Day time is defined as zenith < 93 degrees.
    """  # noqa: E501

    flags = _check_power_limits(
        power, solar_zenith, capacity, capacity_limit_low,
        capacity_limit_high_day, capacity_limit_high_night
        )

    return flags


@mask_flags('LIMITS EXCEEDED')
def check_dc_power_limits(power, solar_zenith, capacity,
                          capacity_limit_low=-0.05,
                          capacity_limit_high_day=1.20,
                          capacity_limit_high_night=0.05):
    """ Checks for extreme AC power.

    Parameters
    ----------
    power : Series
        DC or AC power.
    solar_zenith : Series
        Solar zenith angle in degrees.
    capacity : float
        DC capacity.
    capacity_limit_low : float
        Lower bound in fraction of capacity.
    capacity_limit_high_day : float
        Upper bound in fraction of capacity for day time.
    capacity_limit_high_night : float
        Upper bound in fraction of capacity for night time.

    Returns
    -------
    flags : Series
        True for values that are within the limits:
          * power > capacity * capacity_limit_low, AND
            * power < capacity * capacity_limit_high_day AND solar_zenith < 93, OR
            * power < capacity * capacity_limit_high_night AND solar_zenith > 93

    Notes
    -----
    Day time is defined as zenith < 93 degrees.
    """  # noqa: E501

    flags = _check_power_limits(
        power, solar_zenith, capacity, capacity_limit_low,
        capacity_limit_high_day, capacity_limit_high_night
        )

    return flags


def _check_power_limits(
        power, solar_zenith, capacity, capacity_limit_low,
        capacity_limit_high_day, capacity_limit_high_night
        ):

    # convert fractions to absolute values
    capacity_low = capacity * capacity_limit_low
    capacity_high_day = capacity * capacity_limit_high_day
    capacity_high_night = capacity * capacity_limit_high_night

    # True for daytime values
    day_night = check_irradiance_day_night(solar_zenith, max_zenith=93)

    flag_low = _check_limits(power, lb=capacity_low)
    flag_high_day = _check_limits(power, ub=capacity_high_day) & day_night
    flag_high_night = _check_limits(power, ub=capacity_high_night) & ~day_night

    # composite constructed such that True values within limits for day
    # or night. False values exceed any limit.
    flags = flag_low & (flag_high_day | flag_high_night)
    return flags


@mask_flags('CLEARSKY EXCEEDED')
def check_ghi_clearsky(ghi, ghi_clearsky, kt_max=1.1):
    """
    Flags GHI values greater than clearsky values.

    Parameters
    ----------
    ghi : Series
        Global horizontal irradiance in W/m^2
    ghi_clearsky : Series
         Global horizontal irradiance in W/m^2 under clear sky conditions
    kt_max : float
        maximum clearness index that defines when ghi exceeds clear-sky value.

    Returns
    -------
    flags : Series
        True if ghi is less than or equal to clear sky value.
    """
    kt = clearsky_index(ghi, ghi_clearsky, max_clearsky_index=np.Inf)
    flags = _check_limits(kt, ub=kt_max, ub_le=True)
    return flags


@mask_flags('CLEARSKY EXCEEDED')
def check_poa_clearsky(poa_global, poa_clearsky, kt_max=1.1):
    """
    Flags plane of array irradiance values greater than clearsky values.

    Parameters
    ----------
    poa_global : Series
        Plane of array irradiance in W/m^2
    poa_clearsky : Series
        Plane of array irradiance under clear sky conditions, in W/m^2
    kt_max : float
        maximum allowed ratio of poa_global to poa_clearsky

    Returns
    -------
    flags : Series
        True if poa_global is less than or equal to clear sky value.
    """
    kt = clearsky_index(poa_global, poa_clearsky,
                        max_clearsky_index=np.Inf)
    flags = _check_limits(kt, ub=kt_max, ub_le=True)
    return flags


@mask_flags('NIGHTTIME')
def check_irradiance_day_night(solar_zenith, max_zenith=87):
    """ Checks for day/night periods based on solar zenith.

    Parameters
    ----------
    solar_zenith : Series
        Solar zenith angle in degrees
    max_zenith : maximum zenith angle for a daylight time

    Returns
    -------
    flags : Series
        True when solar zenith is less than max_zenith.
    """
    flags = _check_limits(solar_zenith, ub=max_zenith)
    return flags


@mask_flags('UNEVEN FREQUENCY')
def check_timestamp_spacing(times, freq):
    """ Checks if spacing between times conforms to freq.

    Parameters
    ----------
    times : DatetimeIndex
    freq : string or Timedelta
        Expected frequency of times

    Returns
    -------
    flags : Series
        True when the difference between one time and the time before
        conforms to freq
    """
    if not isinstance(freq, pd.Timedelta):
        freq = pd.Timedelta(freq)
    # first value is NaT, rest are timedeltas
    delta = times.to_series().diff()
    flags = delta == freq
    flags.iloc[0] = True
    return flags


def _all_close_to_first(x, rtol=1e-5, atol=1e-8):
    """ Returns True if all values in x are close to x[0].

    Parameters
    ----------
    x : array
    rtol : float, default 1e-5
        relative tolerance for detecting a change in data values
    atol : float, default 1e-8
        absolute tolerance for detecting a change in data values

    Returns
    -------
    Boolean
    """
    return np.allclose(a=x, b=x[0], rtol=rtol, atol=atol)


@mask_flags('STALE VALUES', invert=False)
def detect_stale_values(x, window=6, rtol=1e-5, atol=1e-8):
    """ Detects stale data.

    For a window of length N, the last value (index N-1) is considered stale
    if all values in the window are close to the first value (index 0).

    Parameters
    ----------
    x : Series
        data to be processed
    window : int, default 6
        number of consecutive values which, if unchanged, indicates stale data
    rtol : float, default 1e-5
        relative tolerance for detecting a change in data values
    atol : float, default 1e-8
        absolute tolerance for detecting a change in data values

    Parameters rtol and atol have the same meaning as in numpy.allclose

    Returns
    -------
    flags : Series
        True if the value is part of a stale sequence of data

    Raises
    ------
        ValueError if window < 2
    """
    if window < 2:
        raise ValueError(f'window set to {window}, must be at least 2')

    flags = x.rolling(window=window).apply(
        _all_close_to_first, raw=True, kwargs={'rtol': rtol, 'atol': atol}
        ).fillna(False).astype(bool)
    return flags


def stale_interpolated_window(interval_length):
    """Returns the recommended window size for detect stale and
    interpolation functions"""
    if interval_length < pd.Timedelta('1h'):
        return 6
    else:
        return 3


@mask_flags('INTERPOLATED VALUES', invert=False)
def detect_interpolation(x, window=6, rtol=1e-5, atol=1e-8):
    """ Detects sequences of data which appear linear.

    Sequences are linear if the first difference appears to be constant.
    For a window of length N, the last value (index N-1) is flagged
    if all values in the window appear to be a line segment.

    Parameters
    ----------
    x : series
        data to be processed
    window : int, default 6
        number of sequential values that, if the first difference is constant,
        are classified as a linear sequence
    rtol : float, default 1e-5
        tolerance relative to max(abs(x.diff()) for detecting a change
    atol : float, default 1e-8
        absolute tolerance for detecting a change in first difference

    Returns
    -------
    flags : Series
        True if the value is part of a linear sequence

    Raises
    ------
        ValueError if window < 3
    """
    if window < 3:
        raise ValueError(f'window set to {window}, must be at least 3')

    # reduce window by 1 because we're passing the first difference
    flags = detect_stale_values(x.diff(periods=1), window=window-1, rtol=rtol,
                                atol=atol)
    return flags


def detect_levels(x, count=3, num_bins=100):
    """ Detects plateau levels in data.

    Parameters
    ----------
    x : Series
        data to be processed
    count : int
        number of plateaus to return
    num_bins : int
        number of bins to use in histogram that finds plateau levels

    Returns
    -------
    levels : list of tuples
        (left, right) values of the interval in x with a detected plateau, in
        decreasing order of count of x values in the interval. List length is
        given by the kwarg count
    """
    nonan = x[~np.isnan(x)]
    hist, bin_edges = np.histogram(nonan, bins=num_bins, density=True)
    level_index = np.argsort(hist * -1)  # decreasing order
    levels = [(bin_edges[i], bin_edges[i + 1]) for i in level_index[:count]]
    return levels, bin_edges


def _label_clipping(x, window, frac):
    """ Returns Series with True at the end of each window with
    sum(x(window)) >= window * frac.
    """
    tmp = x.rolling(window).sum()
    y = (tmp >= window * frac) & x.astype(bool)
    return y


@mask_flags('CLIPPED VALUES', invert=False)
def detect_clipping(ac_power, window=4, fraction_in_window=0.75, rtol=5e-3,
                    levels=2):
    """ Detects clipping in a series of AC power.

    Possible clipped power levels are found by detect_levels. Within each
    sliding window, clipping is indicated when at least fraction_in_window
    of points are close to a clipped power level.

    Parameters
    ----------
    ac_power : Series
        data to be processed

    window : int
        number of data points defining the length of a rolling window

    fraction_in_window : float
        fraction of points which indicate clipping if AC power at each point
        is close to the plateau level

    rtol : float
        a point is close to a clipped level M if
        abs(ac_power - M) < rtol * max(ac_power)

    levels : int
        number of clipped power levels to consider.

    Returns
    -------
    flags : Series
        True when clipping is indicated.
    """
    num_bins = np.ceil(1.0 / rtol).astype(int)
    flags = pd.Series(index=ac_power.index, data=False)
    power_plateaus, bins = detect_levels(ac_power, count=levels,
                                         num_bins=num_bins)
    for lower, upper in power_plateaus:
        temp = pd.Series(index=ac_power.index, data=0.0)
        temp.loc[(ac_power >= lower) & (ac_power <= upper)] = 1.0
        flags = flags | _label_clipping(temp, window=window,
                                        frac=fraction_in_window)
    return flags


@mask_flags('CLEARSKY', invert=False)
def detect_clearsky_ghi(ghi, ghi_clearsky):
    """ Identifies times when GHI is consistent with clear sky conditions.

    Uses the function pvlib.clearsky.detect_clearsky. Assumes ghi data with
    regular (constant) time intervals which must be 15 minutes or less.

    Parameters
    ----------
    ghi : Series
        Global horizontal irradiance in W/m^2

    ghi_clearsky : Series
         Global horizontal irradiance in W/m^2 under clear sky conditions

    Returns
    -------
    flags : Series
        True when clear sky conditions are indicated.

    Notes
    -----
    Clear-sky conditions are inferred when each of six criteria are met; see
    `pvlib.clearsky.detect_clearsky` for references and details. Threshold
    values for each criterion were originally developed for ten minute windows
    containing one-minute data [1]. As indicated in [2], the algorithm also
    works for longer windows and data at different intervals, if threshold
    criteria are roughly scaled to the window length. Here, the threshold
    values are based on [1] with the scaling indicated in [2].

    Warns
    -----
    RuntimeWarning
        If `pvlib.clearsky.detect_clearsky` cannot be applied to the input.

    References
    ----------
    [1] Reno, M.J. and C.W. Hansen, "Identification of periods of clear
    sky irradiance in time series of GHI measurements" Renewable Energy,
    v90, p. 520-531, 2016.

    [2] B. H. Ellis, M. Deceglie and A. Jain, "Automatic Detection of
    Clear-Sky Periods From Irradiance Data," in IEEE Journal of Photovoltaics,
    vol. 9, no. 4, pp. 998-1005, July 2019. doi: 10.1109/JPHOTOV.2019.2914444
    """
    if len(ghi) < 2:
        warnings.warn(
            'At least two datapoints are required for detect_clearsky_ghi',
            RuntimeWarning
        )
        return pd.Series(0, index=ghi.index)
    # determine window length in minutes, 10 x interval for intervals <= 15m
    delta = ghi.index.to_series().diff()
    delta_minutes = delta[1] / np.timedelta64(1, '60s')
    deltas_same = (delta[1:] == delta[1]).all()
    if delta_minutes <= 15 and deltas_same:
        window_length = np.minimum(10*delta_minutes, 60.0)
        scale_factor = window_length / 10
        flags = _detect_clearsky(ghi, ghi_clearsky, ghi.index, window_length,
                                 lower_line_length=-5*scale_factor,
                                 upper_line_length=10*scale_factor,
                                 slope_dev=8*scale_factor)
        return flags
    else:
        warnings.warn(
            'detect_clearsky requires regular time intervals of 15m or less',
            RuntimeWarning
        )
        return pd.Series(0, index=ghi.index)



from pathlib import Path
import shutil


import aiohttp
from asynctest import CoroutineMock, MagicMock
import pandas as pd
from pkg_resources import resource_filename, Requirement
import pytest


from solarforecastarbiter.io.fetch import nwp


def test_gfs_valid_hr_gen():
    expected = (list(range(120)) + list(range(120, 240, 3)) +
                list(range(240, 385, 12)))
    assert list(nwp.GFS_0P25_1HR['valid_hr_gen'](0)) == expected


def test_nam_valid_hr_gen():
    expected = list(range(36)) + list(range(36, 85, 3))
    assert list(nwp.NAM_CONUS['valid_hr_gen'](0)) == expected


@pytest.mark.asyncio
async def test_get_with_retries(mock_sleep):
    getfcn = CoroutineMock()
    getfcn.side_effect = [
        aiohttp.ClientResponseError(MagicMock(), MagicMock())] * 4 + ['hi']
    res = await nwp.get_with_retries(getfcn, retries=5)
    assert res == 'hi'
    assert getfcn.await_count == 5


@pytest.mark.asyncio
async def test_get_with_retries_fails(mock_sleep):
    getfcn = CoroutineMock()
    getfcn.side_effect = [
        aiohttp.ClientResponseError(MagicMock(), MagicMock())] * 4 + ['hi']
    with pytest.raises(aiohttp.ClientResponseError):
        await nwp.get_with_retries(getfcn, retries=4)


@pytest.mark.asyncio
@pytest.mark.parametrize('page,expected,model', [
    ("""
    nam.20190409 other things
    more before ... 201900 nam.20190410
    """, ['20190409', '20190410'], nwp.NAM_CONUS),
    ("""
    gfs.20190409 other things
    more before ... 201900 nam.20190410
    """, ['20190409'], nwp.GFS_0P25_1HR),
])
async def test_get_available_dirs(mocker, page, expected, model):
    get = mocker.patch('solarforecastarbiter.io.fetch.nwp.get_with_retries',
                       new=CoroutineMock())
    get.return_value = page
    assert await nwp.get_available_dirs(None, model) == set(expected)


def test_process_params():
    paramgen = nwp._process_params(nwp.GFS_0P25_1HR,
                                   pd.Timestamp('20190409T1200Z'))
    params = next(paramgen)
    assert 'subregion' in params
    assert 'leftlon' in params
    assert 'endpoint' in params
    assert 'file' in params
    assert 'valid_hr_gen' not in params
    assert 'lev_surface' in params
    assert params['file'] == 'gfs.t12z.pgrb2.0p25.f000'


@pytest.mark.asyncio
async def test_check_next_inittime(mocker):
    session = MagicMock()
    session.head.return_value.__aenter__.return_value.status = 200
    assert await nwp.check_next_inittime(
        session, pd.Timestamp('20190101T0000Z'), nwp.RAP)


@pytest.mark.asyncio
async def test_check_next_inittime_400(mocker):
    session = MagicMock()
    session.head.return_value.__aenter__.return_value.status = 400
    assert not await nwp.check_next_inittime(
        session, pd.Timestamp('20190101T0000Z'), nwp.RAP)


@pytest.mark.asyncio
async def test_check_next_inittime_raises(mocker):
    session = MagicMock()
    session.head.return_value.__aenter__.return_value.status = 200
    session.head.return_value.__aenter__.side_effect = aiohttp.ClientOSError
    assert not await nwp.check_next_inittime(
        session, pd.Timestamp('20190101T0000Z'), nwp.RAP)


@pytest.fixture()
def mock_sleep(mocker):
    sleep = mocker.patch('solarforecastarbiter.io.fetch.nwp.asyncio.sleep',
                         new=CoroutineMock())
    sleep.return_value = None


@pytest.mark.asyncio
async def test_files_to_retrieve(mocker, mock_sleep):
    session = MagicMock()
    session.head.return_value.__aenter__.return_value.headers.__getitem__.return_value = '20190101T0000Z'  # NOQA

    params = [p async for p in nwp.files_to_retrieve(
        session, nwp.RAP, Path('/'), pd.Timestamp('20190409T1200Z'))]
    assert len(params) == 22


@pytest.mark.asyncio
async def test_files_to_retrieve_ends_after_no_new(mocker, mock_sleep):
    def make_header_mocks(init):
        for i in range(100):
            # first time needs headers for lastmod
            if i == 0:
                hmock = MagicMock()
                hmock.headers.__getitem__.return_value = init
                yield hmock
            elif i in range(3, 13):
                yield aiohttp.ClientOSError
            else:
                yield i

    model = nwp.RAP.copy()
    model['avg_max_run_length'] = '30min'
    init = pd.Timestamp('20190409T0000Z')
    headers = list(make_header_mocks(init))
    session = MagicMock()
    session.head.return_value.__aenter__.side_effect = headers
    checknext = mocker.patch(
            'solarforecastarbiter.io.fetch.nwp.check_next_inittime',
            new=CoroutineMock())
    checknext.return_value = False

    params = [p async for p in nwp.files_to_retrieve(session, model, Path('/'),
                                                     init)]

    assert len(params) == 22
    assert checknext.await_count == 10


@pytest.mark.asyncio
async def test_files_to_retrieve_cut_short(mocker, mock_sleep):
    def make_header_mocks(init):
        for i in range(100):
            # first time needs headers for lastmod
            if i == 0:
                hmock = MagicMock()
                hmock.headers.__getitem__.return_value = init
                yield hmock
            elif i == 10:
                yield aiohttp.ClientOSError
            else:
                yield i

    model = nwp.RAP.copy()
    # so now is after avg max run length
    model['avg_max_run_length'] = '30min'
    init = pd.Timestamp('20190409T0000Z')
    headers = list(make_header_mocks(init))
    session = MagicMock()
    session.head.return_value.__aenter__.side_effect = headers
    checknext = mocker.patch(
            'solarforecastarbiter.io.fetch.nwp.check_next_inittime',
            new=CoroutineMock())
    checknext.return_value = True
    params = [p async for p in nwp.files_to_retrieve(session, model, Path('/'),
                                                     init)]
    assert len(params) == 10


@pytest.mark.asyncio
async def test_sleep_until_inittime(mocker):
    m = {'delay_to_first_forecast': '30s'}
    inittime = pd.Timestamp.utcnow()
    sleep = mocker.patch('solarforecastarbiter.io.fetch.nwp.asyncio.sleep',
                         new=CoroutineMock())

    await nwp.sleep_until_inittime(inittime, m)
    assert sleep.mock_calls[0]


@pytest.mark.asyncio
async def test_startup_find_next_runtime(mocker, tmp_path):
    mocker.patch('solarforecastarbiter.io.fetch.nwp.get_available_dirs',
                 new=CoroutineMock(return_value=['20190409',
                                                 '2019041000',
                                                 '2019041006',
                                                 '2019041012']))
    mocker.patch('solarforecastarbiter.io.fetch.nwp.sleep_until_inittime',
                 new=CoroutineMock())
    model = {'update_freq': '6h', 'filename': 'file.nc'}
    withnc = ['2019/04/09/00', '2019/04/09/06', '2019/04/09/12',
              '2019/04/09/18', '2019/04/10/00', '2019/04/10/12']
    for p in withnc:
        pp = tmp_path / p
        pp.mkdir(parents=True)
        (pp / 'file.nc').touch()

    res = await nwp.startup_find_next_runtime(tmp_path, None, model)
    assert res == pd.Timestamp('20190410T0600Z')


@pytest.mark.asyncio
async def test_startup_find_next_runtime_all_there(mocker, tmp_path):
    mocker.patch('solarforecastarbiter.io.fetch.nwp.get_available_dirs',
                 new=CoroutineMock(return_value=['20190409',
                                                 '2019041000',
                                                 '2019041006',
                                                 '2019041012']))
    mocker.patch('solarforecastarbiter.io.fetch.nwp.sleep_until_inittime',
                 new=CoroutineMock())
    model = {'update_freq': '6h', 'filename': 'file.nc'}
    withnc = ['2019/04/09/00', '2019/04/09/06', '2019/04/09/12',
              '2019/04/09/18', '2019/04/10/00', '2019/04/10/06',
              '2019/04/10/12']
    for p in withnc:
        pp = tmp_path / p
        pp.mkdir(parents=True)
        (pp / 'file.nc').touch()

    res = await nwp.startup_find_next_runtime(tmp_path, None, model)
    assert res == pd.Timestamp('20190410T1800Z')


@pytest.mark.asyncio
async def test_startup_find_next_runtime_all_there_out_of_order(
        mocker, tmp_path):
    mocker.patch('solarforecastarbiter.io.fetch.nwp.get_available_dirs',
                 new=CoroutineMock(return_value=['20190409',
                                                 '2019041006',
                                                 '2019041012',
                                                 '2019041000']))
    mocker.patch('solarforecastarbiter.io.fetch.nwp.sleep_until_inittime',
                 new=CoroutineMock())
    model = {'update_freq': '6h', 'filename': 'file.nc'}
    withnc = ['2019/04/09/00', '2019/04/09/06', '2019/04/09/12',
              '2019/04/10/00', '2019/04/10/06', '2019/04/09/18',
              '2019/04/10/12']
    for p in withnc:
        pp = tmp_path / p
        pp.mkdir(parents=True)
        (pp / 'file.nc').touch()

    res = await nwp.startup_find_next_runtime(tmp_path, None, model)
    assert res == pd.Timestamp('20190410T1800Z')


@pytest.mark.asyncio
async def test_next_run_time(mocker, tmp_path):
    model = {'update_freq': '12h', 'filename': 'file.nc'}
    mocker.patch('solarforecastarbiter.io.fetch.nwp.sleep_until_inittime',
                 new=CoroutineMock())
    init = pd.Timestamp('20190409T0000Z')
    withnc = ['2019/04/09/00', '2019/04/09/12',
              '2019/04/10/12']
    for p in withnc:
        pp = tmp_path / p
        pp.mkdir(parents=True)
        (pp / 'file.nc').touch()
    res = await nwp.next_run_time(init, tmp_path, model)
    assert res == pd.Timestamp('20190410T0000Z')


@pytest.mark.asyncio
async def test_optimize_only(mocker, tmp_path):
    if shutil.which('wgrib2') is None:
        pytest.skip('wgrib2 must be on the PATH to run this test')
    grib_dir = tmp_path / 'grib'
    shutil.copytree(
        resource_filename(
            Requirement.parse('solarforecastarbiter'),
            'solarforecastarbiter/io/fetch/tests/data/grib'),
        grib_dir)

    async def run(func, *args, **kwargs):
        return func(*args, **kwargs)

    mocker.patch('solarforecastarbiter.io.fetch.signal.pthread_kill')
    mocker.patch('solarforecastarbiter.io.fetch.nwp.run_in_executor',
                 new=run)
    await nwp.optimize_only(grib_dir, 'rap')
    assert [x.name for x in grib_dir.iterdir()] == ['rap.nc']



import pytest
import re
import pandas as pd
from requests.exceptions import HTTPError

from solarforecastarbiter.datamodel import Site, Observation
from solarforecastarbiter.io import api
from solarforecastarbiter.io.reference_observations import eia


@pytest.fixture
def session(requests_mock):
    return api.APISession('')


@pytest.fixture
def site():
    return Site(
        name='CAISO',
        latitude=37.0,
        longitude=-120.0,
        elevation=0.0,
        timezone='Etc/GMT+7',
        site_id='',
        provider='',
        extra_parameters='{"network_api_id": "CISO-ALL", "attribution": "https://www.eia.gov/opendata/", "network": "EIA", "network_api_abbreviation": "eia", "observation_interval_length": 1}',  # noqa: E501
    )


@pytest.fixture
def site_no_extra():
    return Site(
        name='CAISO',
        latitude=37.0,
        longitude=-120.0,
        elevation=0.0,
        timezone='Etc/GMT+7',
        site_id='',
        provider='',
        extra_parameters='',
    )


@pytest.fixture
def log(mocker):
    log = mocker.patch('solarforecastarbiter.io.reference_observations.'
                       'eia.logger')
    return log


@pytest.fixture()
def mock_list_sites(mocker, many_sites):
    mocker.patch('solarforecastarbiter.io.api.APISession.list_sites',
                 return_value=many_sites)


def test_initialize_site_observations(
        requests_mock, mocker, session, site, single_observation,
        single_observation_text, mock_list_sites):
    matcher = re.compile(f'{session.base_url}/observations/.*')
    requests_mock.register_uri('POST', matcher,
                               text=single_observation.observation_id)
    requests_mock.register_uri('GET', matcher, content=single_observation_text)
    status = mocker.patch(
        'solarforecastarbiter.io.api.APISession.create_observation')
    eia.initialize_site_observations(session, site)
    assert status.called


@pytest.fixture
def mock_obs_creation(mocker):
    mocked = mocker.patch(
        'solarforecastarbiter.io.reference_observations.common.'
        'create_observation')
    return mocked


@pytest.fixture
def mock_obs_creation_error(mocker, mock_obs_creation):
    mock_obs_creation.side_effect = HTTPError(
        response=mocker.Mock(response=mocker.Mock(text="eror")))


def test_initialize_site_obs(mock_api, mock_obs_creation, site):
    eia.initialize_site_observations(mock_api, site)
    mock_obs_creation.assert_called()


def test_initialize_site_obs_http_error(
        log, mock_api, mock_obs_creation_error, site):
    eia.initialize_site_observations(mock_api, site)
    assert log.error.call_count == 1
    assert log.debug.call_count == 1


def test_fetch(mocker, session, site):
    status = mocker.patch(
        'solarforecastarbiter.io.fetch.eia.get_eia_data'
    )
    start = pd.Timestamp('2020-01-01T0000Z')
    end = pd.Timestamp('2020-01-02T0000Z')
    api_key = 'bananasmoothie'

    index = pd.date_range(start, end, freq="1h")
    data = {"net_load": range(len(index))}
    df = pd.DataFrame(index=index, data=data)
    status.return_value = df

    out = eia.fetch(session, site, start, end, eia_api_key=api_key)
    assert status.called
    assert not out.empty


def test_fetch_empty(log, mocker, session, site):
    status = mocker.patch(
        'solarforecastarbiter.io.fetch.eia.get_eia_data'
    )
    start = pd.Timestamp('2020-01-01T0000Z')
    end = pd.Timestamp('2020-01-02T0000Z')
    api_key = 'bananasmoothie'
    status.return_value = pd.DataFrame()
    out = eia.fetch(session, site, start, end, eia_api_key=api_key)
    assert status.called
    assert log.warning.call_count == 1
    assert out.empty


def test_fetch_fail(mocker, session, site_no_extra):
    start = pd.Timestamp('2020-01-01T0000Z')
    end = pd.Timestamp('2020-01-02T0000Z')
    api_key = 'bananasmoothie'
    out = eia.fetch(session, site_no_extra, start, end, eia_api_key=api_key)
    assert out.empty


def test_initialize_site_forecasts(mocker, session, site, mock_list_sites):

    obs = Observation(
        name='Net Load',
        variable='net_load',
        site=site,
        interval_label='ending',
        interval_value_type='interval_mean',
        interval_length=pd.Timedelta('1h'),
        uncertainty=0.0,
    )

    mocker.patch('solarforecastarbiter.io.api.APISession.list_observations',
                 return_value=[obs])
    mocker.patch('solarforecastarbiter.io.api.APISession.list_forecasts')
    mocker.patch('solarforecastarbiter.io.api.APISession.'
                 'list_probabilistic_forecasts')
    status = mocker.patch(
        'solarforecastarbiter.io.api.APISession.create_forecast')
    eia.initialize_site_forecasts(session, site)
    assert status.called


@pytest.fixture
def mock_eia_creds(mocker):
    mocker.patch.dict('os.environ', {'EIA_API_KEY': 'fake_key'})


def test_update_observation_data(mocker, session, site, mock_eia_creds):
    obs_update = mocker.patch(
        'solarforecastarbiter.io.reference_observations.common.'
        'update_site_observations')
    start = pd.Timestamp('20200101T0000Z')
    end = pd.Timestamp('20200102T0000Z')
    eia.update_observation_data(session, [site], [], start, end)
    obs_update.assert_called()
    assert obs_update.call_count == 1


def test_update_observation_data_no_creds(session, site):
    start = pd.Timestamp('20200101T0000Z')
    end = pd.Timestamp('20200102T0000Z')
    with pytest.raises(KeyError) as e:
        eia.update_observation_data(session, [site], [], start, end)
    assert 'environment variable' in str(e.value)



"""Collection of Functions to convert API responses into python objects
and vice versa.
"""
from contextlib import contextmanager
from functools import wraps
from inspect import signature
import json
import re


import pandas as pd


def _dataframe_to_json(payload_df):
    payload_df.index.name = 'timestamp'
    json_vals = payload_df.tz_convert("UTC").reset_index().to_json(
        orient="records", date_format='iso', date_unit='s')
    return '{"values":' + json_vals + '}'


def observation_df_to_json_payload(
        observation_df, default_quality_flag=None):
    """Extracts a variable from an observation DataFrame and formats it
    into a JSON payload for posting to the Solar Forecast Arbiter API.

    Parameters
    ----------
    observation_df : DataFrame
        Dataframe of observation data. Must contain a tz-aware DateTimeIndex
        and a 'value' column. May contain a column of data quality
        flags labeled 'quality_flag'.
    default_quality_flag : int
        If 'quality_flag' is not a column, the quality flag for each row is
        set to this value.

    Returns
    -------
    string
        SolarForecastArbiter API JSON payload for posting to the observation
        endpoint. See Notes section for example.

    Notes
    -----
    Function returns an object in the following format:

    .. code::

        {
          'values': [
            {
              “timestamp”: “2018-11-22T12:01:48Z”, # ISO 8601 datetime in UTC
              “value”: 10.23, # floating point value of observation
              “quality_flag”: 0
            },...
          ]
        }

    Raises
    ------
    KeyError
       When 'value' is missing from the columns or 'quality_flag'
       is missing and default_quality_flag is None
    """
    if default_quality_flag is None:
        payload_df = observation_df[['value', 'quality_flag']]
    else:
        payload_df = observation_df[['value']]
        payload_df['quality_flag'] = int(default_quality_flag)
    return _dataframe_to_json(payload_df)


def forecast_object_to_json(forecast_series):
    """
    Converts a forecast Series to JSON to post to the
    SolarForecastArbiter API.

    Parameters
    ----------
    forecast_series : pandas.Series
        The series that contains the forecast values with a
        datetime index.

    Returns
    -------
    string
        The JSON encoded forecast values dict
    """
    payload_df = forecast_series.to_frame('value')
    return _dataframe_to_json(payload_df)


def _json_to_dataframe(json_payload):
    # in the future, might worry about reading the response in chunks
    # to stream the data and avoid having it all in memory at once,
    # but 30 days of 1 minute data is probably ~4 MB of text. A better
    # approach would probably be to switch to a binary format.
    vals = json_payload['values']
    if len(vals) == 0:
        df = pd.DataFrame([], columns=['value', 'quality_flag'],
                          index=pd.DatetimeIndex([], name='timestamp',
                                                 tz='UTC'))
    else:
        df = pd.DataFrame.from_dict(json_payload['values'])
        df.index = pd.to_datetime(df['timestamp'], utc=True,
                                  infer_datetime_format=True)
    return df


def json_payload_to_observation_df(json_payload):
    """
    Convert the JSON payload dict as returned by the SolarForecastArbiter API
    observations/values endpoint into a DataFrame

    Parameters
    ----------
    json_payload : dict
        Dictionary as returned by the API with a "values" key which is a list
        of dicts like {'timestamp': <timestamp>, 'value': <float>,
        'quality_flag': <int>}

    Returns
    -------
    pandas.DataFrame
       With a tz-aware DatetimeIndex and ['value', 'quality_flag'] columns
       and dtypes {'value': float, 'quality_flag': int}
    """
    df = _json_to_dataframe(json_payload)
    return df[['value', 'quality_flag']].astype(
        {'value': float, 'quality_flag': int})


def json_payload_to_forecast_series(json_payload):
    """
    Convert the JSON payload dict as returned by the SolarForecastArbiter API
    forecasts/values endpoing into a Series

    Parameters
    ----------
    json_payload : dict
        Dictionary as returned by the API with a "values" key which is a list
        of dicts like {'timestamp': <timestamp>, 'value': <float>}

    Returns
    -------
    pandas.Series
       With a tz-aware DatetimeIndex and float dtype
    """

    df = _json_to_dataframe(json_payload)
    return df['value'].astype(float)


def adjust_start_end_for_interval_label(interval_label, start, end,
                                        limit_instant=False):
    """
    Adjusts the start and end times depending on the interval_label.

    Parameters
    ----------
    interval_label : str or None
       The interval label for the the object the data represents
    start : pandas.Timestamp
       Start time to restrict data to
    end : pandas.Timestamp
       End time to restrict data to
    limit_instant : boolean
       If true, an interval label of 'instant' will remove a nanosecond
       from end to ensure forecasts do not overlap. If False, instant
       returns start, end unmodified

    Returns
    -------
    start, end
       Return the adjusted start and end

    Raises
    ------
    ValueError
       If an invalid interval_label is given

    Examples
    --------
    .. testsetup::

       from solarforecastarbiter.io.utils import *

    Define input start/end:

    >>> start = pd.Timestamp('20190101 1200Z')
    >>> end = pd.Timestamp('20190101 1300Z')

    Beginning:

    >>> adjust_start_end_for_interval_label('beginning', start, end)
    (Timestamp('2019-01-01 12:00:00+0000', tz='UTC'), Timestamp('2019-01-01 12:59:59.999999999+0000', tz='UTC'))

    Ending:

    >>> adjust_start_end_for_interval_label('ending', start, end)
    (Timestamp('2019-01-01 12:00:00.000000001+0000', tz='UTC'), Timestamp('2019-01-01 13:00:00+0000', tz='UTC'))

    Instantaneous:

    >>> adjust_start_end_for_interval_label('instant', start, end)
    (Timestamp('2019-01-01 12:00:00+0000', tz='UTC'), Timestamp('2019-01-01 13:00:00+0000', tz='UTC'))

    >>> adjust_start_end_for_interval_label('instant', start, end,
    ...                                     limit_instant=True)
    (Timestamp('2019-01-01 12:00:00+0000', tz='UTC'), Timestamp('2019-01-01 12:59:59.999999999+0000', tz='UTC'))

    """ # NOQA

    if (
            interval_label is not None and
            interval_label not in ('instant', 'beginning', 'ending')
    ):
        raise ValueError('Invalid interval_label')

    if (
            interval_label == 'beginning' or
            (interval_label == 'instant' and limit_instant)
    ):
        end -= pd.Timedelta(1, unit='nano')
    elif interval_label == 'ending':
        start += pd.Timedelta(1, unit='nano')
    return start, end


def adjust_timeseries_for_interval_label(data, interval_label, start, end):
    """
    Adjusts the index of the data depending on the interval_label, start,
    and end. Will always return the data located between start, end.

    Parameters
    ----------
    data : pandas.Series or pandas.DataFrame
       The data with a localized DatetimeIndex
    interval_label : str or None
       The interval label for the the object the data represents
    start : pandas.Timestamp
       Start time to restrict data to
    end : pandas.Timestamp
       End time to restrict data to

    Returns
    -------
    pandas.Series or pandas.DataFrame
       Return data between start and end, in/excluding the endpoints
       depending on interval_label

    Raises
    ------
    ValueError
       If an invalid interval_label is given or data is not localized.
    """
    start, end = adjust_start_end_for_interval_label(interval_label, start,
                                                     end)
    data = data.sort_index(axis=0)
    # pandas >= 0.25.1 requires start, end to have same tzinfo.
    # unexpected behavior when data is not localized, so prevent that
    if data.empty:
        return data
    if data.index.tzinfo is None:
        raise ValueError('data must be localized')
    start = start.tz_convert(data.index.tzinfo)
    end = end.tz_convert(data.index.tzinfo)
    return data.loc[start:end]


def serialize_timeseries(ser):
    """
    Serialize a timeseries to JSON. Note that the microseconds
    portion of the index will be discarded.

    Parameters
    ----------
    ser : {pandas.Series, pandas.DataFrame}
       Must have a tz-localized datetime index

    Returns
    -------
    str
        The JSON serialized data along with a schema

    Raises
    ------
    TypeError
        If the input is invalid
    """
    if not (
            isinstance(ser, (pd.Series, pd.DataFrame)) and
            isinstance(ser.index, pd.DatetimeIndex) and
            ser.index.tzinfo is not None
    ):
        raise TypeError(
            'Only pandas Series or DataFrame with a localized DatetimeIndex '
            'is supported')
    v = ser.copy()
    v.index.name = 'timestamp'
    if isinstance(v, pd.Series):
        jsonvals = v.tz_convert('UTC').reset_index(name='value').to_json(
            orient='records', date_format='iso', date_unit='s')
        column = 'value'
        dtype = str(v.dtype)
        objtype = 'Series'
    else:
        v.index.name = 'timestamp'
        jsonvals = v.tz_convert('UTC').reset_index().to_json(
            orient='records', date_format='iso', date_unit='s')
        column = v.columns.astype(str).to_list()
        dtype = v.dtypes.astype(str).to_list()
        objtype = 'DataFrame'
    schema = {
        'version': 1,
        'orient': 'records',
        'timezone': 'UTC',
        'column': column,
        'index': 'timestamp',
        'dtype': dtype,
        'objtype': objtype
    }
    out = '{"schema":' + json.dumps(schema) + ',"data":' + jsonvals + '}'
    return out


def deserialize_timeseries(data):
    """
    Deserializes a timeseries from JSON

    Parameters
    ----------
    data : str
        JSON string to deserialize. Must have schema and data keys.

    Returns
    -------
    pandas.Series or pandas.DataFrame
        Deserialized timeseries

    Raises
    ------
    ValueError
       If "schema" or "data" keys are not found in the JSON string
    KeyError
       If the schema object does not contain the proper keys
    """
    schema_str = re.search('(?<="schema":)\\s*{[^{}]*}\\s*(?=(,|}))', data)
    if schema_str is None:
        raise ValueError('Could not locate schema in data string')
    schema = json.loads(schema_str.group(0))
    if schema['version'] == 0:
        # compatibility with data serialized and stored before the
        # objtype key was added to schema and DataFrames were suppored in v1
        objtype_str = 'Series'
    else:
        objtype_str = schema['objtype']
    # find between "data": and , or }, with only one set of []
    data_str = re.search('(?<="data":)\\s*\\[[^\\[\\]]*\\](?=\\s*(,|}))', data)
    if data_str is None:
        raise ValueError('Could not locate data key in data string')
    df = pd.read_json(data_str.group(0), orient=schema['orient'],
                      convert_dates=True)
    if objtype_str == 'Series':
        if df.empty:
            return pd.Series([], name=schema['column'], index=pd.DatetimeIndex(
                [], tz=schema['timezone'], name='timestamp'),
                            dtype=schema.get('dtype', float))
        out = df.set_index(schema['index'])[schema['column']].astype(
            schema['dtype'])
    elif objtype_str == 'DataFrame':
        if df.empty:
            try:
                dtype = schema['dtype'][0]
            except IndexError:
                dtype = float
            return pd.DataFrame(
                [], columns=schema['column'], index=pd.DatetimeIndex(
                    [], tz=schema['timezone'], name='timestamp'), dtype=dtype)
        out = df.set_index(schema['index'])
        # pd.read_json will set all column names to strings, so
        # columns originally specified with float names need to be
        # mapped back into the right name dtype. this code is not needed
        # if columns are always strings.
        # str_col_map = {str(col): col for col in schema['column']}
        # out = out.rename(columns=str_col_map)
        out = out.astype(dict(zip(schema['column'], schema['dtype'])))
    if out.index.tzinfo is None:
        out = out.tz_localize(schema['timezone'])
    return out


class HiddenToken:
    """
    Obscure the representation of the input string `token` to avoid saving
    or displaying access tokens in logs.
    """
    def __init__(self, token):
        self.token = str(token)  # make sure it isn't a localproxy

    def __repr__(self):
        return '****ACCESS*TOKEN****'


def ensure_timestamps(*time_args):
    """
    Decorator that converts the specified time arguments of the wrapped
    function to pandas.Timestamp objects

    Parameters
    ----------
    strings
       Function arguments to convert to pandas.Timestamp before
       executing function

    Raises
    ------
    ValueError
        If any of time_args cannot be converted to pandas.Timestamp

    Examples
    --------
    .. testsetup::

       import datetime as dt
       from solarforecastarbiter.io.utils import *

    >>> @ensure_timestamps('start', 'end')
    ... def get_values(start, end, other_arg):
    ...     # do stuff with start, end assumed to be pandas.Timestamps
    ...     if isinstance(start, pd.Timestamp):
    ...         return True

    >>> get_values('2019-01-01T00:00Z', dt.datetime(2019, 1, 2, 12), 'other')
    True
    """
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            sig = signature(f)
            inds = {k: None for k in time_args}
            for i, k in enumerate(sig.parameters.keys()):
                if k in inds:
                    inds[k] = i
            nargs = list(args)
            for k, ind in inds.items():
                if k in kwargs:
                    kwargs[k] = pd.Timestamp(kwargs[k])
                elif ind is not None:
                    nargs[ind] = pd.Timestamp(args[ind])
            return f(*nargs, **kwargs)
        return wrapper
    return decorator


def load_report_values(raw_report, values):
    """
    Load the processed forecast/observation data into the
    datamodel.ProcessedForecastObservation objects of the raw_report.

    Parameters
    ----------
    raw_report : datamodel.RawReport
        The raw report with processed_forecasts_observations to
        be replaced
    values : list
        The report values dict as returned by the API.

    Returns
    -------
    tuple
       Of datamodel.ProcessedForecastObservation with values loaded into
       `forecast_values` and `observation_values`
    """
    val_dict = {v['id']: v['processed_values'] for v in values}
    out = []
    for fxobs in raw_report.processed_forecasts_observations:
        fx_vals = val_dict.get(fxobs.forecast_values, None)
        if fx_vals is not None:
            fx_vals = deserialize_timeseries(fx_vals)
        obs_vals = val_dict.get(fxobs.observation_values, None)
        if obs_vals is not None:
            obs_vals = deserialize_timeseries(obs_vals)
        ref_fx_vals = val_dict.get(fxobs.reference_forecast_values)
        if ref_fx_vals is not None:
            ref_fx_vals = deserialize_timeseries(ref_fx_vals)
        new_fxobs = fxobs.replace(forecast_values=fx_vals,
                                  observation_values=obs_vals,
                                  reference_forecast_values=ref_fx_vals)
        out.append(new_fxobs)
    return tuple(out)


@contextmanager
def mock_raw_report_endpoints(base_url):
    """
    Mock API report endpoints under base_url to enable testing
    of the report generation task run via the dashboard. Requires
    requests_mock>=1.8.0. Catches all report endpoints, but if
    report generation requires POSTing to other endpoints in the
    future, they will need to be added here.
    """
    import requests_mock
    value_dict = {}  # for raw processed values

    def post_value_callback(request, context):
        context.status_code = 200
        rjson = request.json()
        id_ = rjson['object_id']
        value_dict[id_] = rjson['processed_values']
        return id_

    raw_dict = {}  # for the raw reports

    def post_raw_callback(request, context):
        context.status_code = 200
        raw_dict.update(request.json())

    report_dict = {}  # for new and full reports

    def post_report_callback(request, context):
        context.status_code = 200
        report_dict.update(request.json())
        return 'no_id'

    def get_report_callback(request, context):
        context.status_code = 200
        out = report_dict.copy()
        if raw_dict:
            out['raw_report'] = raw_dict
            out['values'] = [
                {'id': k, 'processed_values': v}
                for k, v in value_dict.items()]
        out['status'] = 'complete'
        return out

    with requests_mock.Mocker(real_http=True) as m:
        value_re = re.compile(f'{base_url}/reports/.*/values')
        raw_re = re.compile(f'{base_url}/reports/.*/raw')
        m.register_uri('GET', re.compile(
            f'{base_url}/reports/.*'),
                       json=get_report_callback)
        m.register_uri('POST', re.compile(
            f'{base_url}/reports/*'),
                       text=post_report_callback)
        m.register_uri('POST', re.compile(
            f'{base_url}/reports/.*/status/.*'))
        m.register_uri('POST', value_re, text=post_value_callback)
        m.register_uri('POST', raw_re, json=post_raw_callback)
        yield



import pytest
from solarforecastarbiter.metrics import event


@pytest.mark.parametrize("obs,fx,result", [
    # one event
    ([True], [False], (0, 0, 0, 1)),
    ([False], [True], (0, 1, 0, 0)),
    ([True], [True], (1, 0, 0, 0)),
    ([False], [False], (0, 0, 1, 0)),

    # two events
    ([True, False], [True, False], (1, 0, 1, 0)),
    ([True, True], [True, False], (1, 0, 0, 1)),
    ([False, False], [True, False], (0, 1, 1, 0)),
    ([True, True], [True, True], (2, 0, 0, 0)),
    ([False, False], [False, False], (0, 0, 2, 0)),
    ([True, True], [False, False], (0, 0, 0, 2)),

    # bad data
    pytest.param([True], [], 0,
                 marks=pytest.mark.xfail(raises=RuntimeError, strict=True)),
    pytest.param([], [True], 0,
                 marks=pytest.mark.xfail(raises=RuntimeError, strict=True)),
    pytest.param([True], [True, True], 0,
                 marks=pytest.mark.xfail(raises=RuntimeError, strict=True))
])
def test__event2count(obs, fx, result):
    tp, fp, tn, fn = event._event2count(obs, fx)
    assert (tp, fp, tn, fn) == result
    assert (tp + fp + tn + fn) == len(obs)


@pytest.mark.parametrize("obs,fx,result", [
    ([True], [False], 0.0),
    ([False], [True], 0.0),
    ([True, True], [True, True], 1.0),
    ([True, True], [True, False], 0.5),
])
def test_probability_of_detection(obs, fx, result):
    assert event.probability_of_detection(obs, fx) == result


@pytest.mark.parametrize("obs,fx,result", [
    ([True], [False], 0.0),
    ([False], [True], 1.0),
    ([True, True], [True, False], 0.0),
    ([False, True], [True, True], 0.5),
])
def test_false_alarm_ratio(obs, fx, result):
    assert event.false_alarm_ratio(obs, fx) == result


@pytest.mark.parametrize("obs,fx,result", [
    ([True], [False], 0.0),
    ([False], [True], 1.0),
    ([False, False], [True, False], 0.5),
])
def test_probability_of_false_detection(obs, fx, result):
    assert event.probability_of_false_detection(obs, fx) == result


@pytest.mark.parametrize("obs,fx,result", [
    ([True], [False], 0.0),
    ([True], [True], 1.0),
    ([False], [False], 0.0),
    ([True, False], [True, True], 0.5),
    ([True, False, True], [True, True, False], 1 / 3),
])
def test_critical_success_index(obs, fx, result):
    assert event.critical_success_index(obs, fx) == result


@pytest.mark.parametrize("obs,fx,result", [
    ([True], [False], 0.0),
    ([False], [False], 0.0),
    ([True, True], [True, False], 0.5),
    ([True, False], [True, True], 2.0),
])
def test_event_bias(obs, fx, result):
    assert event.event_bias(obs, fx) == result


@pytest.mark.parametrize("obs,fx,result", [
    ([True], [False], 0.0),
    ([False], [True], 0.0),
    ([True, True], [True, True], 1.0),
    ([True, False], [True, False], 1.0),
    ([True, False], [True, True], 0.5),
])
def test_event_accuracy(obs, fx, result):
    assert event.event_accuracy(obs, fx) == result



import inspect
import os


import pandas as pd


from solarforecastarbiter.io.fetch import pvdaq

TEST_DATA_DIR = os.path.dirname(
    os.path.abspath(inspect.getfile(inspect.currentframe())))
SYSTEM_FILE = os.path.join(TEST_DATA_DIR, 'data', 'pvdaq_metadata.json')
DATA_FILE = os.path.join(TEST_DATA_DIR, 'data', 'pvdaq_2020_data.csv')
DATA_FILE_2019 = os.path.join(TEST_DATA_DIR, 'data', 'pvdaq_2019_data.csv')


def test_get_pvdaq_metadata(requests_mock):
    with open(SYSTEM_FILE) as f:
        content = f.read()

    requests_mock.register_uri(
        'GET', 'https://developer.nrel.gov/api/pvdaq/v3/sites.json',
        content=content.encode()
    )

    expected = [{
        'available_years': [2010,
                            2011,
                            2012,
                            2013,
                            2014,
                            2016,
                            2017,
                            2018,
                            2019,
                            2020],
        'comments': None,
        'confidential': False,
        'inverter_mfg': 'Fronius',
        'inverter_model': 'IG 4000',
        'module_mfg': 'Sharp',
        'module_model': 'ND-208U1',
        'module_tech': 2,
        'name_private': 'Nelson Home, Golden, CO, Array 1',
        'name_public': '[2] Residential #1a',
        'site_area': 22.82,
        'site_azimuth': 181.2,
        'site_elevation': 1675.0,
        'site_latitude': 39.7214,
        'site_longitude': 105.0972,
        'site_power': 2912.0,
        'site_tilt': 18.5,
        'system_id': 2}]
    out = pvdaq.get_pvdaq_metadata(2, 'doesntmatter')
    assert out == expected


def test_get_pvdaq_data(requests_mock):
    with open(DATA_FILE) as f:
        content = f.read()

    requests_mock.register_uri(
        'GET', 'https://developer.nrel.gov/api/pvdaq/v3/data_file',
        content=content.encode()
    )

    data = pvdaq.get_pvdaq_data(1276, 2020, 'fakekey')
    assert data.index[0] == pd.Timestamp('2020-01-01 00:00:00')
    assert data.index[-1] == pd.Timestamp('2020-01-02 23:45:00')
    some_cols = ['ac_current', 'ac_power', 'dc_power', 'wind_speed']
    for col in some_cols:
        assert col in data.columns


def test_get_pvdaq_data_2years(requests_mock):

    with open(DATA_FILE) as f:
        content = f.read()

    requests_mock.register_uri(
        'GET',
        'https://developer.nrel.gov/api/pvdaq/v3/data_file?api_key=fakekey&system_id=1276&year=2020',  # noqa: E501
        content=content.encode()
    )

    with open(DATA_FILE_2019) as f:
        content_2019 = f.read()

    requests_mock.register_uri(
        'GET',
        'https://developer.nrel.gov/api/pvdaq/v3/data_file?api_key=fakekey&system_id=1276&year=2019',  # noqa: E501
        content=content_2019.encode()
    )

    data = pvdaq.get_pvdaq_data(1276, [2019, 2020], 'fakekey')
    assert data.index[0] == pd.Timestamp('2019-01-01 00:00:00')
    assert data.index[-1] == pd.Timestamp('2020-01-02 23:45:00')
    some_cols = ['ac_current', 'ac_power', 'dc_power', 'wind_speed']
    for col in some_cols:
        assert col in data.columns
